@@ -1,6 +1,6 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2012-2019  Ruby-GNOME Project Team
+ *  Copyright (C) 2012-2021  Ruby-GNOME Project Team
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -2011,11 +2011,18 @@ rb_gi_return_argument_to_ruby(GICallableInfo *callable_info,
     GITypeInfo return_value_info;
 
     may_return_null = g_callable_info_may_return_null(callable_info);
+    g_callable_info_load_return_type(callable_info, &return_value_info);
     if (may_return_null && !argument->v_pointer) {
-        return Qnil;
+        GITypeTag return_value_tag = g_type_info_get_tag(&return_value_info);
+        switch (return_value_tag) {
+          case GI_TYPE_TAG_GLIST:
+          case GI_TYPE_TAG_GSLIST:
+            return rb_ary_new();
+          default:
+            return Qnil;
+        }
     }
 
-    g_callable_info_load_return_type(callable_info, &return_value_info);
     rb_argument = rb_gi_argument_to_ruby(argument, FALSE, &return_value_info,
                                          in_args, out_args, args_metadata);
     switch (g_callable_info_get_caller_owns(callable_info)) {
