@@ -1,4 +1,4 @@
-# Copyright (C) 2012-2019  Ruby-GNOME2 Project Team
+# Copyright (C) 2012-2021  Ruby-GNOME Project Team
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -80,6 +80,10 @@ module GObjectIntrospection
       define_singleton_method(@base_module, name, info)
     end
 
+    def load_function_info_singleton_method(info, klass, method_name)
+      define_singleton_method(klass, method_name, info)
+    end
+
     def define_module_function(target_module, name, function_info)
       function_info.unlock_gvl = should_unlock_gvl?(function_info, target_module)
       full_method_name = "#{target_module}\#.#{name}"
@@ -125,7 +129,7 @@ module GObjectIntrospection
         end
         unless method_infos.empty?
           base_class = @base_module.const_get(base_class_name)
-          load_method_infos(method_infos, base_class.singleton_class)
+          load_methods_method(method_infos, base_class.singleton_class)
         end
       else
         return if info.gtype_struct?
@@ -212,6 +216,7 @@ module GObjectIntrospection
       klass = self.class.define_class(info.gtype,
                                       rubyish_class_name(info),
                                       @base_module)
+      load_virtual_functions(info, klass)
       load_fields(info, klass)
       load_methods(info, klass)
     end
@@ -277,11 +282,11 @@ module GObjectIntrospection
         next if method_infos.empty?
         case method_infos.first
         when ConstructorInfo
-          load_constructor_infos(method_infos, klass)
+          load_methods_constructor(method_infos, klass)
         when MethodInfo
-          load_method_infos(method_infos, klass)
+          load_methods_method(method_infos, klass)
         when FunctionInfo
-          load_function_infos(method_infos, klass)
+          load_methods_function(method_infos, klass)
         else
           raise "TODO: #{method_info_class}"
         end
@@ -292,16 +297,18 @@ module GObjectIntrospection
       end
     end
 
-    def load_constructor_infos(infos, klass)
+    def load_methods_constructor(infos, klass)
       return if infos.empty?
 
       call_initialize_post = lambda do |object|
         initialize_post(object)
       end
+      invokers = []
       infos.each do |info|
         name = "initialize_#{info.name}"
         info.unlock_gvl = should_unlock_gvl?(info, klass)
         invoker = Invoker.new(info, name, "#{klass}\##{name}")
+        invokers << invoker
         klass.__send__(:define_method, name) do |*arguments, &block|
           invoker.invoke(self, arguments, block)
           call_initialize_post.call(self)
@@ -309,57 +316,46 @@ module GObjectIntrospection
         klass.__send__(:private, name)
       end
 
-      find_info = lambda do |arguments|
-        find_suitable_callable_info(infos, arguments)
+      initialize = lambda do |receiver, arguments, block|
+        invokers.each do |invoker|
+          catch do |tag|
+            invoker.invoke(receiver, arguments.dup, block, tag)
+            call_initialize_post.call(receiver)
+            return
+          end
+        end
+        message = "wrong arguments: "
+        message << "#{klass.name}#initialize("
+        message << arguments.collect(&:inspect).join(", ")
+        message << "): "
+        message << "available signatures"
+        invokers.each do |invoker|
+          message << ": #{invoker.signature}"
+        end
+        raise ArgumentError, message
       end
       klass.__send__(:define_method, "initialize") do |*arguments, &block|
-        info = find_info.call(arguments, &block)
-        __send__("initialize_#{info.name}", *arguments, &block)
+        initialize.call(self, arguments, block)
       end
     end
 
+    def load_virtual_functions(info, klass)
+      klass.extend(VirtualFunctionImplementable)
+      gtype_prefix = rubyish_gtype_name(klass.gtype.name)
+      implementor = VirtualFunctionImplementor.new(self.class,
+                                                   gtype_prefix,
+                                                   info.vfuncs)
+      klass.__send__(:initialize_virtual_function_implementable,
+                     implementor)
+    end
+
+    def rubyish_gtype_name(name)
+      name.scan(/[A-Z]+[a-z\d]+/).collect(&:downcase).join("_")
+    end
+
     def initialize_post(object)
     end
 
-    def find_suitable_callable_info(infos, arguments)
-      min_n_args = nil
-      max_n_args = nil
-      candidate_infos = []
-      infos.each do |info|
-        n_in_args = info.n_in_args
-        n_required_in_args = info.n_required_in_args
-        if (n_required_in_args..n_in_args).cover?(arguments.size)
-          candidate_infos << info
-        end
-        min_n_args = [min_n_args || n_required_in_args, n_required_in_args].min
-        max_n_args = [max_n_args || n_in_args, n_in_args].max
-      end
-
-      if candidate_infos.size == 1
-        return candidate_infos.first
-      elsif candidate_infos.size > 1
-        candidate_info = candidate_infos.find do |info|
-          in_arg_infos = info.in_args
-          arguments.each.with_index.all? do |argument, i|
-            match_argument?(in_arg_infos[i], argument)
-          end
-        end
-        return candidate_info || candidate_infos.first
-      end
-
-      detail = "#{arguments.size} for #{min_n_args}"
-      if min_n_args < max_n_args
-        detail << "..#{max_n_args}"
-      end
-      raise ArgumentError, "wrong number of arguments (#{detail})"
-    end
-
-    def match_argument?(arg_info, argument)
-      return true if argument.nil? and arg_info.may_be_null?
-
-      arg_info.type.match?(argument)
-    end
-
     def rubyish_method_name(function_info, options={})
       name = function_info.name
       if options[:prefix]
@@ -506,7 +502,7 @@ module GObjectIntrospection
       false
     end
 
-    def load_method_infos(infos, klass)
+    def load_methods_method(infos, klass)
       infos.each do |info|
         method_name = rubyish_method_name(info)
         load_method_info(info, klass, method_name)
@@ -550,12 +546,12 @@ module GObjectIntrospection
       end
     end
 
-    def load_function_infos(infos, klass)
+    def load_methods_function(infos, klass)
       infos.each do |info|
         name = rubyish_method_name(info)
         next if name == "new"
         next if name == "alloc"
-        define_singleton_method(klass, name, info)
+        load_function_info_singleton_method(info, klass, name)
       end
     end
 
@@ -565,6 +561,7 @@ module GObjectIntrospection
         self.class.define_interface(info.gtype,
                                     rubyish_class_name(info),
                                     @base_module)
+      load_virtual_functions(info, interface_module)
       load_methods(info, interface_module)
     end
 
@@ -591,7 +588,7 @@ module GObjectIntrospection
         @prepared = false
       end
 
-      def invoke(receiver, arguments, block)
+      def invoke(receiver, arguments, block, abort_tag=nil)
         ensure_prepared
 
         if receiver and @function_info_p
@@ -599,13 +596,16 @@ module GObjectIntrospection
         end
 
         arguments, block = build(receiver, arguments, block)
-        unless valid?(arguments)
-          if @on_invalid == :fallback
-            return @value_on_invalid
+        if wrong_number_of_arguments?(arguments)
+          if abort_tag
+            throw(abort_tag)
           else
             raise ArgumentError, invalid_error_message(arguments)
           end
         end
+        unless normalize_arguments!(arguments, abort_tag)
+          return @value_on_invalid
+        end
 
         if block.nil? and @require_callback_p
           receiver.to_enum(@method_name, *arguments)
@@ -623,6 +623,12 @@ module GObjectIntrospection
         end
       end
 
+      def signature
+        ensure_prepared
+        argument_signatures = @in_args.collect(&:signature)
+        "(" + argument_signatures.join(", ") + ")"
+      end
+
       private
       def ensure_prepared
         return if @prepared
@@ -684,12 +690,26 @@ module GObjectIntrospection
         return arguments, block
       end
 
-      def valid?(arguments)
-        return false unless @valid_n_args_range.cover?(arguments.size)
-        if @on_invalid == :fallback
-          arguments.zip(@in_arg_types) do |argument, type|
-            return false unless type.match?(argument)
+      def wrong_number_of_arguments?(arguments)
+        not @valid_n_args_range.cover?(arguments.size)
+      end
+
+      def normalize_arguments!(arguments, abort_tag)
+        arguments.size.times do |i|
+          argument = arguments[i]
+          type = @in_arg_types[i]
+          converted_argument = type.try_convert(argument)
+          if converted_argument.nil?
+            next if argument.nil?
+            if abort_tag
+              throw(abort_tag)
+            elsif @on_invalid == :fallback
+              return false
+            else
+              next
+            end
           end
+          arguments[i] = converted_argument
         end
         true
       end
@@ -704,5 +724,53 @@ module GObjectIntrospection
         "#{@full_method_name}: wrong number of arguments (#{detail})"
       end
     end
+
+    class VirtualFunctionImplementor
+      def initialize(loader_class, gtype_prefix, infos)
+        @loader_class = loader_class
+        @gtype_prefix = gtype_prefix
+        @infos = {}
+        prefix = GLib::VIRTUAL_FUNCTION_IMPLEMENTATION_PREFIX
+        infos.each do |info|
+          name = info.name
+          @infos[:"#{prefix}#{name}"] = info
+          @infos[:"#{prefix}#{gtype_prefix}_#{name}"] = info
+        end
+      end
+
+      def implement(implementor_gtype, name)
+        info = @infos[name]
+        return false if info.nil?
+        container = info.container
+        vtable_gtype = container.gtype
+        if container.respond_to?(:class_struct)
+          struct = container.class_struct
+        else
+          return false unless implementor_gtype.type_is_a?(vtable_gtype)
+          struct = container.iface_struct
+        end
+        field = struct.find_field(info.name)
+        @loader_class.implement_virtual_function(field,
+                                                 implementor_gtype,
+                                                 vtable_gtype,
+                                                 name.to_s)
+        true
+      end
+    end
+
+    module VirtualFunctionImplementable
+      def initialize_virtual_function_implementable(implementor)
+        @virtual_function_implementor = implementor
+      end
+
+      def implement_virtual_function(implementor_class, name)
+        unless instance_variable_defined?(:@virtual_function_implementor)
+          return false
+        end
+        @virtual_function_implementor.implement(implementor_class.gtype,
+                                                name)
+        true
+      end
+    end
   end
 end
